#ifdef WIN32
#pragma warning ( disable : 4786 )
#pragma warning ( disable : 4800 )
#include "windows.h"
#endif

#include "my_sql.h"
#include <mysql.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <mysql_com.h>
#include <cmath>

#ifndef WIN32
#include <sys/time.h>
#include <unistd.h> 
#endif

#include <strstream>

#include "change_type.h"
#include "specials.h"
#include "stringhelper.h"

#ifndef rcsid
static const char rcsid[]="$Id: my_sql.cpp,v 2.51 2001/04/23 22:48:30 ral Exp $";
#endif

const char* CMySql::MessageArray[]={
    "row(s) affected",              //0
    "row(s) found",
    " WHERE ",
    "No table given to work on",
    "Not connected",
    "I do not change system database in this way", //5
    "No database given to work on",
};

int CMySql::MessageCount = 5;

const my_sql_type CMySql::f_arr[]=
{   //                                                             def  aut   zer   uns  key   keys  enums binary
    my_sql_type("TINYINT","TinyInt (1 Byte)",          1,  3,0,0, true, true, true, true,false, true,false,false),
    my_sql_type("SMALLINT","SmallInt (2 Bytes)",       1,  5,0,0, true, true, true, true,false, true,false,false),
    my_sql_type("MEDIUMINT","MediumInt (3 Bytes)",     1,  8,0,0, true, true, true, true,false, true,false,false),
    my_sql_type("INT","Int (4 Bytes)",                 1, 10,0,0, true, true, true, true,false, true,false,false),
    my_sql_type("BIGINT","BigInt (8 Bytes)",           1, 20,0,0, true, true, true, true,false, true,false,false),
    my_sql_type("FLOAT","Float (4 Bytes)",             1, 20,0,4, true, true, true,false,false, true,false,false),
    my_sql_type("REAL","Double (8 Bytes)",             1, 20,0,8, true, true, true,false,false, true,false,false),
    my_sql_type("DECIMAL","Decimal (lenght,dec)",      1, 20,0,8, true, true, true,false, true, true,false,false),
    my_sql_type("CHAR","Char (1 <= num <=255)",        1,255,0,0,false,false,false,false, true, true,false, true),
    my_sql_type("VARCHAR","Varchar (1 <= num <=255)",  1,255,0,0,false,false,false,false, true, true,false, true),
    my_sql_type("TINYBLOB","Tinyblob (num <= 255)",    2,  2,0,0,false,false,false,false,false,false,false,false), //10
    my_sql_type("BLOB","Blob (num <= 65535)",          2,  2,0,0,false,false,false,false,false,false,false,false),
    my_sql_type("MEDIUMBLOB","Mediumblob (num < 2^24)",2,  2,0,0,false,false,false,false,false,false,false,false),
    my_sql_type("LONGBLOB","Longblob (num < 2^32)",    2,  2,0,0,false,false,false,false,false,false,false,false),
    my_sql_type("TINYTEXT","Tinytext (num <= 255)",    2,  2,0,0,false,false,false,false,false,false,false,false),
    my_sql_type("TEXT","Text (num <= 65535)",          2,  2,0,0,false,false,false,false,false,false,false,false),
    my_sql_type("MEDIUMTEXT","Mediumtext (num < 2^24)",2,  2,0,0,false,false,false,false,false,false,false,false),
    my_sql_type("LONGTEXT","Longtext (num < 2^32)",    2,  2,0,0,false,false,false,false,false,false,false,false),
    my_sql_type("TIMESTAMP","Timestamp (4 Bytes)",     1, 14,0,0, true,false,false,false,false, true,false,false),
    my_sql_type("DATE","Date (4 Bytes)",              10, 10,0,0, true,false,false,false,false, true,false,false),
    my_sql_type("TIME","Time (3 Bytes)",               8,  8,0,0, true,false,false,false,false, true,false,false), //20
    my_sql_type("ENUM","Enum",                         1, 80,0,0, true,false,false,false,false, true, true,false),
    my_sql_type("SET","Set",                           1, 80,0,0, true,false,false,false,false, true, true,false),
    my_sql_type("DATETIME","DateTime",                14, 14,0,0,false,false,false,false,false,false,false,false),
    my_sql_type("YEAR","Year",                         4,  4,0,0, true,false,false,false,false,false,false,false),
    my_sql_type("DOUBLE","Double (8 Bytes)",           1, 20,0,8, true, true, true,false,false, true,false,false),
    my_sql_type("NUMERIC","Decimal (lenght,dec)",      1, 20,0,8, true, true, true,false, true, true,false,false),
};

const my_sql_type&CMySql::sql_type(int i)
{
    return f_arr[i];
}

int CMySql::FieldDescriptCounts()
{
    return sizeof(f_arr)/sizeof(my_sql_type);
}

const char* CMySql::FieldDescript(int i)
{
    if (i >= FieldDescriptCounts() || i < 0)
        return "";
    return f_arr[i]._descriptor.c_str();
}

const char* CMySql::TypeDescript(int i)
{
    if (i >= FieldDescriptCounts() || i < 0) 
        return "";
    return f_arr[i]._name.c_str();
}

bool CMySql::is_system_db(const char*name)
{
    if (!name || strlen(name)==0) {
	return false;
    }
    return (strcasecmp(name,"mysql")==0);
}

void CMySql::sql_disconnect()
{
    nothing_else = 1;
    if (Sql_Connection) {
        mysql_close(Sql_Connection);
        delete Sql_Connection;
        Sql_Connection = 0L;
    }
    Server_MasterVersion = Server_MinorVersion = Server_PatchLevel = 0;
    nothing_else = 0;
}

CMySql::CMySql()
    :refCounter()
{
    Sql_Connection = 0L;
    nothing_else = 0;
    Server_MasterVersion = Server_MinorVersion = Server_PatchLevel = 0;
    m_Host = m_Pass = m_User = m_Port = "";
}

CMySql::CMySql(const CMySql&old)
    :refCounter()
{
    Sql_Connection = 0L;
    nothing_else = 0;
    Server_MasterVersion = Server_MinorVersion = Server_PatchLevel = 0;
    m_Host = m_Pass = m_User = m_Port = "";
    message_call_back = old.message_call_back;
    sql_connect(old.m_Host.c_str(),old.m_User.c_str(),
		old.m_Pass.c_str(),old.m_Port.c_str());
}

CMySql::~CMySql()
{
    sql_disconnect();
}

void CMySql::sql_connect(const std::string&host,const std::string&user,
			 const std::string&pass,const std::string&port)
{
    sql_connect(host.c_str(),user.c_str(),pass.c_str(),port.c_str());
}

void CMySql::sql_connect(const char*host,const char*user,const char*pass,const char*port)
{
    int Port;
    nothing_else = 1;
    
    Port = port?atoi(port):3306;
    m_Port = port?port:"";
    m_User = user?user:"";
    m_Pass = pass?pass:"";
    m_Host = host?host:"";

    sql_disconnect();
    char*s_version = NULL;
    char*help_ptr = NULL;
    Server_MasterVersion = Server_MinorVersion = Server_PatchLevel = 0;
    Sql_Connection = new MYSQL;
    if (!Sql_Connection)
        return;
    mysql_init(Sql_Connection);
    MYSQL * dummy_pointer = mysql_real_connect(Sql_Connection,host,user,pass,(char*)0,(unsigned int)Port,(char*)0,0);
    if (dummy_pointer) {
        do_message("Connected");
        s_version = mysql_get_server_info(Sql_Connection);
        if (s_version) {
            Server_MasterVersion = atoi(s_version);
            help_ptr = strstr(s_version,".");
            if (help_ptr) {
                help_ptr = &help_ptr[1];
                Server_MinorVersion = atoi(help_ptr);
            }
            help_ptr = strstr(help_ptr,".");
            if (help_ptr) {
                help_ptr = &help_ptr[1];
                Server_PatchLevel = atoi(help_ptr);
            }
        }
    } else {
	do_message(FormatSTLString("Couldn't connect: %s",mysql_error(Sql_Connection)));
	delete Sql_Connection;
	Sql_Connection = 0;
    }
    nothing_else = 0;
}

int CMySql::refresh_grants()
{
    return do_refresh(REFRESH_GRANT,"grants reloaded");
}

int CMySql::refresh_tables()
{
    return do_refresh(REFRESH_TABLES,"tables flushed");
}

int CMySql::refresh_cache()
{
    return do_refresh(REFRESH_HOSTS,"cache flushed");
}

int CMySql::refresh_log()
{
    return do_refresh(REFRESH_LOG,"new log started");
}

int CMySql::do_refresh(int What, const char*ok_msg)
{
    int i = 0;
    if (!Sql_Connection) {
        do_message("not connected");
        return -1;
    }
    i = mysql_refresh(Sql_Connection, What);
    if (i < 0)
        do_message(get_last_error());
    else
        do_message(ok_msg);
    return i;
}

const std::string&CMySql::get_info()
{
    char*c_info;
    char*h_info;
    int p_info;
    char*s_info;
    char t_1[] = "Clientversion";
    if (nothing_else) {
        connect_info = "";
    } else {
        if (Sql_Connection) {
            c_info = mysql_get_client_info();
            h_info = mysql_get_host_info(Sql_Connection);
            p_info = mysql_get_proto_info(Sql_Connection);
            s_info = mysql_get_server_info(Sql_Connection);
#ifndef WIN32
            connect_info = FormatSTLString("%s\n%s %s\nProtokoll Version: %i\nServerversion: %s",
		    h_info,t_1,c_info,p_info,s_info);
#else
            connect_info = FormatSTLString("%s\r\n%s %s\r\nProtokoll Version: %i\r\nServerversion: %s",
		    h_info,t_1,c_info,p_info,s_info);
#endif
        }
        else {
            connect_info = "Couldn't get info";
        }
    }
    return connect_info;
}

char*CMySql::get_status()
{
    char * Message;
    char * Fehler;
    if (nothing_else || !Sql_Connection)
        return 0;
    Message = mysql_stat(Sql_Connection);
    Fehler = get_last_error();
    if (Fehler[0] != 0)
        do_message(Fehler);
    return Message;
}

bool CMySql::get_threads(stringlist&target)const
{
    std::string result;
    MYSQL_RES*Res;
    MYSQL_ROW row;
    if (!Sql_Connection || nothing_else)
        return true;
    Res=mysql_list_processes(Sql_Connection);
    const char*st;
    st = mysql_error(Sql_Connection);
    if (st[0] != 0) {
        do_message(st);
    }
    if (!Res)
        return false;
    result = "";
    while ((row=mysql_fetch_row(Res))) {
        for (unsigned int i = 0;i < mysql_num_fields(Res);++i) {
            result+=row[i]!=0?(char*)row[i]:" ";
            if(i < mysql_num_fields(Res) - 1)
                result+="\t";
        }
        target.push_back(result);
        result = "";
    }
    mysql_free_result(Res);
    return true;
}

bool CMySql::get_threads(list_vector&target)const
{
    MYSQL_RES*Res;
    MYSQL_ROW row;
    target.resize(0);
    if (!Sql_Connection || nothing_else)
        return true;
    Res = mysql_list_processes(Sql_Connection);
    const char*st;
    st = mysql_error(Sql_Connection);
    if (st[0] != 0) {
        do_message(st);
    }
    if (!Res)
        return false;
    stringlist result;
    while((row=mysql_fetch_row(Res))) {
        for (unsigned int i = 0; i < mysql_num_fields(Res);++i) {
            result.push_back(row[i]!=0?row[i]:"");
        }
        target.push_back(result);
        result.clear();
    }
    mysql_free_result(Res);
    return true;
}

/*
* retrieves the list of databases in the current
* connected server
*/
bool CMySql::get_db_list(const char*pattern,stringlist&target)const
{
    target.resize(0);
    std::string result;
    
    MYSQL_RES*Res;
    MYSQL_ROW row;
    if (!Sql_Connection || nothing_else)
        return false;
    Res=mysql_list_dbs(Sql_Connection,pattern);
    const char*st;
    st = mysql_error(Sql_Connection);
    if (st[0] != 0) {
        do_message(st);
    }
    if (!Res)
        return false;
    while ((row=mysql_fetch_row(Res))) {
        for (unsigned int i = 0;i < mysql_num_fields(Res);i++) {
            result+=row[i]?(char*)row[i]:" ";
            if(i < mysql_num_fields(Res) - 1)
                result+="\t";
        }
        target.push_back(result);
        result = "";
    }
    mysql_free_result(Res);
    return true;
}

/*
 * fetches the tables of a database
 */
bool CMySql::get_table_list(const char*pattern,const char*db,stringlist&target)
{
    std::string s;
    MYSQL_RES*Res;
    MYSQL_ROW row;
    target.resize(0);
    if (!Sql_Connection || nothing_else || db==0)
        return false;
    if (connect_db(db) < 0)
        return false;
    const char*st;
    Res = mysql_list_tables(Sql_Connection,pattern);
    st = mysql_error(Sql_Connection);
    if (st[0] != 0) {
        do_message(st);
    }
    if (!Res)
        return false;
    while ((row=mysql_fetch_row(Res))) {
        for (unsigned int i = 0;i < mysql_num_fields(Res);i++) {
            s+=row[i]?(char*)row[i]:"";
            if(i < mysql_num_fields(Res) - 1)
                s+="\t";
        }
        target.push_back(s);
        s = "";
    }
    mysql_free_result(Res);
    return true;
}

/*
 * get the columns of a table
 */
bool CMySql::get_table_fields(const char*db,const char*table,ColumnList&target)const
{
    MYSQL_RES*Res;
    MYSQL_FIELD*Field;
    MYSQL_ROW row;
    CColumnDescriptor entry,empty_descriptor;
    
    target.resize(0);
    std::string command = "";
    Field = 0;
    if (!Sql_Connection || nothing_else || db==0 || table==0)
        return false;

    if (!db) {
		do_message(MessageArray[6]);
		return false;
    }
    if (connect_db(db)!=0) {
        do_message("Could not connect to database");
        return false;
    }

    command = "SHOW COLUMNS FROM ";
    command+=table;
    int i = mysql_query(Sql_Connection,command.c_str());
    if (i < 0) {
        do_message(get_last_error());
        return false;
    }

    Res = mysql_store_result(Sql_Connection);
    if (!Res) {
        //autch - this should never be...
        command = mysql_error(Sql_Connection);
        if (!command.empty())
            do_message(command.c_str());
        else
            do_message("Error fetch the result");
        return 0;
    }
    int f_type = 0;
    unsigned int i_count;
    while ((row=mysql_fetch_row(Res))) {
	entry.set_name((const char*)row[0]);
	std::string type_string = "";
	for (i_count = 0; i_count < strlen((const char*)row[1]);++i_count) {
	    if (row[1][i_count]=='(') {
		break;
	    }
	    type_string +=(char)toupper(row[1][i_count]);
	}
	for (f_type = 0; f_type < FieldDescriptCounts();++f_type) {
	    if (type_string == TypeDescript(f_type)) {
		entry.set_FieldType(f_type);
		break;
	    }
	}
	if (f_type == FieldDescriptCounts()) {
		do_message("Error retrieving fieldtypes - please send a mail to the "
				"autor of the software!");
		return false;
	}
	my_sql_type t_info = sql_type(f_type);
	if (f_type > 10 && f_type < 18) {
	    entry.set_Blob(true);
	}
	if (row[1][i_count] == '(') {
	    //some more type_infos
	    ++i_count;
	    if (f_type == 21 || f_type == 22) {
		++i_count;
		unsigned int max = 0;
		type_string = "";
		for (;i_count < strlen((const char*)row[1]);++i_count) {
		    if (row[1][i_count] == '\'') {
			max = max>type_string.size()?max:type_string.size();
			entry.insert_enum(type_string);
			type_string = "";
			i_count+=2;
		    } else if (row[1][i_count]=='(') {
			break;
		    } else {
			type_string+=row[1][i_count];
		    }
		}
		entry.set_MaxLength(max);
	    } else {
		entry.set_MaxLength(atoi((const char*)&row[1][i_count]));
		char * komma = strstr((const char*)&row[1][i_count],",");
		if (komma) {
		    komma = &komma[1];
		    entry.set_MaxDecimals(atoi(komma));
		}
	    }
	} else {
	    //its only the max value
	    entry.set_MaxLength(t_info._dec_max);
	}
	if (strstr((const char*)row[1],"unsigned")) {
	    entry.set_Unsigned(true);
	} else {
	    entry.set_Unsigned(false);
	}
	if (strstr((const char*)row[1],"zerofill")) {
	    entry.set_ZeroFill(true);
	} else {
	    entry.set_ZeroFill(false);
	}
	if (strstr((const char*)row[1],"binary")) {
	    entry.set_Binary(true);
	} else {
	    entry.set_Binary(false);
	}
	if (row[2] && strstr((const char*)row[2],"YES")) {
	    entry.set_NotNull(false);
	} else {
	    entry.set_NotNull(true);
	}
	if (row[3]) {
	    if (strstr((const char*)row[3],"PRI") ) {
		entry.set_Primary(true);
	    } else if (strstr((const char*)row[3],"UNI")) {
		entry.set_Unique(true);
	    } else if (strstr((const char*)row[3],"MUL")) {
		entry.set_PartOfMultiKey(true);
	    }
	}
	if (row[4]) {
	    entry.set_default((const char*)row[4]);
	}
	if (row[5]) {
	    if (strstr((const char*)row[5],"auto_increment")) {
		entry.set_AutoIncrement(true);
	    }
	}
	target.push_back(entry);
	entry = empty_descriptor;
    }
    mysql_free_result(Res);
    return true;
}

int CMySql::kill_thread(unsigned long pid)
{
    int i;
    if (!Sql_Connection)
        return -1;
    i = mysql_kill(Sql_Connection,pid);
    if (i < 0)
        do_message(get_last_error());
    return i;
}

char*CMySql::get_last_error()const
{
    if (!Sql_Connection) {
        return "No connection";
    }
    return mysql_error(Sql_Connection);
}

/*
 * inserts a new database
 */
int CMySql::create_db(const char*db)
{
    int i = 0;
    const char*f;
    if (!db || !Sql_Connection || nothing_else)
        return -1;
    i = mysql_create_db(Sql_Connection,db);
    f = get_last_error();
    if (i != 0) {
        do_message(f);
        return i;
    }
    return i;
}

bool CMySql::create_table(const char*db,const char*name,ColumnList*columns,KeyList*keys)
{
    std::string statement;
    std::string temp_statement;
	
    ColumnList_iterator desc = NULL;
    KeyList_iterator key = NULL;
    unsigned int count = 0;
    std::string Type;

    if (!name || strlen(name) == 0) {
        do_message(MessageArray[3]);
        return false;
    }
    if (!columns || columns->size() == 0){
        do_message("Wich columns should I create?");
        return false;
    }
    if (!Sql_Connection) {
        do_message(MessageArray[4]);
        return false;
    }
    
    if (connect_db(db)!=0) {
        do_message("Could not connect to database");
        return false;
    }

    if (!db) {
	do_message(MessageArray[6]);
	return false;
    }
    statement = "CREATE TABLE ";
    //statement+=db;statement+='.';
    statement+= name;statement+=" (";

    for (desc = columns->begin(); desc != columns->end(); ++desc) {
        if (desc != columns->begin()) {
            statement+=", ";
        }
        count++;
	if (!build_command(temp_statement,*desc)) {
	    return false;
	}
	statement+=temp_statement;
    }
    if (keys) {
        int already_prim = 0;
        pairIlist col_list;
        
        int f_count = 0;
        for (key = keys->begin(); key != keys->end(); ++key) {
            f_count = 0;
            statement+=", ";
            switch(key->getType()) {
            case CKeyDescriptor::PRIMARY:
                if (!already_prim) {
                    statement+="PRIMARY KEY ";
                    already_prim = 1;
                } else {
                    statement+="UNIQUE ";
                    key->setType(CKeyDescriptor::UNIQUE);
                }
                break;
            case CKeyDescriptor::UNIQUE:
                statement+="UNIQUE ";
                break;
            case CKeyDescriptor::INDEX:
                statement+="INDEX ";
                break;
            default:
                continue;
            }
            if (key->getType() != CKeyDescriptor::PRIMARY) {
                statement+=key->get_name();
            }
            col_list = key->get_ColList();
            if (col_list.size() == 0)
                continue;
            statement+="(";
            pairIlist::iterator iter;
            iter = col_list.begin();
            for (;iter!=col_list.end();++iter) {
                if (f_count> 0)
                    statement+=",";
                ++f_count;
                statement+=iter->first.c_str();
                if (iter->second > 0) {
                    statement+="(";
                    statement+= FormatSTLString("%u",iter->second);
                    statement+=")";
                }
            }
            statement+=") ";
        }
    }
    statement+=")";
    do_message(statement);
    int i = mysql_query(Sql_Connection,statement.c_str());
    if (i < 0) {//error??
        //why?
        do_message(get_last_error());
        return false;
    }
    return true;
}

int CMySql::_fetch_rows(const char*db,const char*table,const char*sort_by,
                   stringlist*which,unsigned int offset,unsigned int count,
                   PatternList*Pattern,bool errors_only,MYSQL_RES**Res)
{
    std::string query;
    std::string entry;
    std::string escaped;
#ifndef WIN32
    timeval starttime,endtime;
#endif

    int i = 0;

    if (!which) {do_message("Which rows?");return -1;}
    if (!Sql_Connection) {do_message(MessageArray[4]);return -1;}
    if (!table || strlen(table) == 0) {do_message(MessageArray[3]);return -1;}
    if (!db) {do_message(MessageArray[6]);return -1;}
    if (connect_db(db)<0) {return -1;}

    query = "";
    stringlist::iterator w_iter = which->begin();
    for (;w_iter!=which->end();++w_iter) {
        if (query.empty())
            query = *w_iter;
        else {
            query+=",";
            query+=*w_iter;
        }
    }
    if (query.empty())
        query+="SELECT *";
    else
        query="SELECT " + query;
    query+=" FROM ";
    //query+=db;query+=".";
    query+=table;query+=" ";
    PatternIterator p_iter;
    if (Pattern && Pattern->size()) {
        p_iter = Pattern->begin();
        //    patter = Pattern->first();
        query+=MessageArray[2];
        //    query+=" ";
        query+=p_iter->get_entry();query+=" LIKE \'";
        do_escape(p_iter->get_pattern(),escaped);
        if (escaped.size()==0)
            return -1;
        query+=escaped;
        query+="\' ";
        ++p_iter;
        for (;p_iter != Pattern->end();++p_iter){
            query+="AND ";query+=p_iter->get_entry();query+=" LIKE \'";
            do_escape(p_iter->get_pattern(),escaped);
            if (escaped.size()==0)
                return -1;
            query+=escaped;
            query+="\' ";
        }
    }
    if (sort_by) {
        query+="ORDER BY ";query+=sort_by;query+=" ";
    }
    if (count == 0)
        count = 99999;
    if (count > 0 || offset > 0) {
        std::string ls;
        std::string cs;
        if (offset > 0) {
            ls = FormatSTLString("LIMIT %u",offset);
            if (count > 0) {
                ls+=",";
            }
            else
                ls+=",-1";
        }
        else {
            ls = "LIMIT ";
        }
        if (count > 0) {
            ls+= FormatSTLString("%u",count);
        }
        query+=ls;
    }
    //we'll never need them again
    which->clear();
    if (Pattern) 
        Pattern->clear();
    //what is the resulting query?
    if (!errors_only) do_message(query.c_str());
#ifndef WIN32    
    gettimeofday(&starttime,0);
#endif
    i = mysql_query(Sql_Connection,query.c_str());

    if (i < 0) {//error??
        //why?
	if (errors_only)do_message(query);
        do_message(get_last_error());
        return -1;
    }
    *Res = mysql_store_result(Sql_Connection);
    if (!Res) {
        //autch - this should never be...
        query = mysql_error(Sql_Connection);
        if (!query.empty())
            do_message(query.c_str());
        else
            do_message("Error fetch the result");
        return -1;
    }
#ifndef WIN32
    gettimeofday(&endtime,0);
    long int sec,usec;
    sec = endtime.tv_sec - starttime.tv_sec;
    usec = endtime.tv_usec - starttime.tv_usec;
    usec += 1000*1000*sec;
    printf("Fetchtime: %li,%06li sec ",usec/(1000*1000),usec%(1000*1000));
#endif
    return 1;
}

/*
 * Makes a simple search
 */
int CMySql::fetch_rows(const char*db,const char*table,const char*sort_by,
                       stringlist*which,unsigned int offset,unsigned int count,
                       PatternList*Pattern,bool errors_only)
{
    MYSQL_RES*Res;
    int i = _fetch_rows (db,table,sort_by,which,offset,count,Pattern,errors_only,&Res);
    if (i < 0)
        return -1;
    unsigned j = 0;
    MYSQL_ROW row;
    MYSQL_FIELD*Field;

    //what did the server found?
    unsigned long zaehler = Res->row_count;
    if (!errors_only) do_message(FormatSTLString("%lu %s",zaehler,MessageArray[1]));
    std::string entry,query;
    entry = "";query = "";
    std::string helper = "";
    int groesse;
#ifndef WIN32
    timeval starttime,endtime;
    long int sec,usec;
    gettimeofday(&starttime,0);
#endif
    while ((row=mysql_fetch_row(Res))) {
        for (j = 0;j < mysql_num_fields(Res);++j) {
            query = row[j]?(char*)row[j]:"";
            //how long is the longest?
            Res->current_field = j;
            Field = mysql_fetch_field(Res);
            // we retrieve the first line splitted, for request, which will only need this line
            // for instance a query to the user rights
            if (which->size() == 0) {
                /*                if (sizes) {
                    sizes->push_back(FormatSTLString("%i",Field->max_length));
                }
                */
                if ( Pattern ) {
                    Pattern->push_back(CPatternPair(Field->name,query));
                }
            }
            //ok - justify
            helper.resize(0);
            if (IS_NUM(Field->type)) {
		groesse = Field->max_length-query.size();
		helper.resize(groesse>=0?groesse:0,' ');
            } else {
                helper = "";
            }
	    helper+=query;
            if (helper.size() == 0)
                helper = " ";
            if (j<mysql_num_fields(Res)-1)
                // its the simplest way to select datas - hope there is no tab in a field
                helper+="\t";
            entry+=helper;
        }
        //append to the list
        which->push_back(entry);
        entry = "";
    }
#ifndef WIN32
    gettimeofday(&endtime,0);
    sec = endtime.tv_sec - starttime.tv_sec;
    usec = endtime.tv_usec - starttime.tv_usec;
    usec+=1000*1000*sec;
    printf("Listbuildtime: %li,%06li sec\n",usec/(1000*1000),usec%(1000*1000));
#endif
    mysql_free_result(Res);
    
    return 0;
}

/*
 * Makes a simple search
 */
int CMySql::fetch_rows(const char*db,const char*table,const char*sort_by,
                       stringlist*which,list_vector&target,unsigned int offset,unsigned int count,
                       PatternList*Pattern,bool errors_only)
{
    MYSQL_RES*Res;
    int i = _fetch_rows (db,table,sort_by,which,offset,count,Pattern,errors_only,&Res);
    if (i < 0)
        return -1;
    unsigned j = 0;
    MYSQL_ROW row;
    MYSQL_FIELD*Field;

    //what did the server found?
    unsigned long zaehler = Res->row_count;
    if (!errors_only) do_message(FormatSTLString("%lu %s",zaehler,MessageArray[1]));
    std::string query;
    query = "";
    stringlist helper,entry;
#ifndef WIN32
    timeval starttime,endtime;
    long int sec,usec;
    gettimeofday(&starttime,0);
#endif
    target.resize(0);
    entry.resize(0);
    while ((row=mysql_fetch_row(Res))) {
        for (j = 0;j < mysql_num_fields(Res);++j) {
            query = row[j]?(char*)row[j]:"";
            //how long is the longest?
            Res->current_field = j;
            Field = mysql_fetch_field(Res);
            // we retrieve the first line splitted, for request, which will only need this line
            // for instance a query to the user rights
            if (which->size() == 0) {
                /*                if (sizes) {
                    sizes->push_back(FormatSTLString("%i",Field->max_length));
                    }*/
                if ( Pattern ) {
                    Pattern->push_back(CPatternPair(Field->name,query));
                }
            }
            entry.push_back(query);
        }
        //append to the list
        target.push_back(entry);
        entry.resize(0);
    }
#ifndef WIN32
    gettimeofday(&endtime,0);
    sec = endtime.tv_sec - starttime.tv_sec;
    usec = endtime.tv_usec - starttime.tv_usec;
    usec+=1000*1000*sec;
    printf("Listbuildtime: %li,%06li sec\n",usec/(1000*1000),usec%(1000*1000));
#endif
    mysql_free_result(Res);
    
    return 0;
}


/*
 * inserts a single row into a table
 */
int CMySql::insert_row(const char*db,const char*table, PatternList*Pattern)
{
    std::string query = "";
    std::string value_string = "";
    std::string row_string = "";

    std::string escaped;
    int i;
    if (!Sql_Connection) {
        do_message(MessageArray[4]);
        return -1;
    }
    if (Pattern == 0 || Pattern->empty()) {
        do_message("No data to insert");
        return -1;
    }
    if (table == 0 || strlen(table) == 0) {
        do_message(MessageArray[3]);
        return -1;
    }
    if (!db)
	return -1;
    i = connect_db(db);
    if (i < 0)
	return i;
    //build the pattern-string
    PatternIterator p_iter;
    for (p_iter = Pattern->begin();p_iter!=Pattern->end();++p_iter){
        if (value_string.empty()) {
            value_string = "\'";
            do_escape(p_iter->get_pattern(),escaped);
            if (!escaped.size())
                return -1;
            value_string += escaped;
            value_string+="\'";
            row_string = p_iter->get_entry();
        }
        else {
            value_string+=",\'";
            do_escape(p_iter->get_pattern(),escaped);
            if (!escaped.size())
                return -1;
            value_string+=escaped;
            value_string+="\'";
            row_string+=",";row_string+=p_iter->get_entry();
        }
    }
    //build the complete query
    query = "INSERT INTO ";
    //query+=db;query+=".";
    query+=table;query+=" (";query+=row_string;query+=") VALUES (";
    query+= value_string;query+=")";
    //send it to the protocoll
    do_message(query.c_str());
    //do it
    i = mysql_query(Sql_Connection,query.c_str());
    if (i == -1) {
        //if some goes wrong 
        query = mysql_error(Sql_Connection);
        do_message(query.c_str());
        return i;
    }
    //should be > 0 
    i = mysql_affected_rows(Sql_Connection);
    do_message(FormatSTLString("%i %s",i,MessageArray[0]));
    return i;
}

std::string&CMySql::do_escape(const std::string&what,std::string&target)
{
    target.resize(0);
    string _s = "%%";
    string _r = "\\%";

    if (what.size()==0)
        return target;

    char * escaped = new char[what.size()*2+1];
    if (!escaped) {
        do_message("Error allocating memory");
        return target;
    }
    memset(escaped,'\0',what.size()*2+1);
    mysql_escape_string(escaped,what.c_str(),what.size());
    target = escaped;

    string::size_type it = target.find(_s);

    while (it!=string::npos) {
        target.replace(it,1,_r);
        cout << "target: " << target.c_str()<< endl;
        ++it;++it;++it;
        it = target.find(_s,it);
    }
    if (escaped)
        delete[]escaped;
    return target;
}

int CMySql::update_row(const char*db,const char*table,
                       PatternList*to_change,
                       PatternList*identifiers)
{
    std::string request;
    std::string clean;
    int i;

    if (!Sql_Connection) {
        do_message(MessageArray[4]);
        return -1;
    }
    if (!table || strlen(table) == 0) {
        do_message(MessageArray[3]);
        return -1;
    }
    if (!to_change || to_change->size() == 0) {
        do_message("Nothing given to change");
        return -1;
    }
    if (!db) {
	do_message(MessageArray[6]);
	return -1;
    }
    i = connect_db(db);
    if (i < 0)
	return i;
    request = "UPDATE ";
    //request+=db;request+=".";
    request+= table;request+=" SET ";
    PatternIterator p_iter;
    for (p_iter = to_change->begin();p_iter != to_change->end();++p_iter) {
        if (p_iter != to_change->begin())
            request+=",";
        request+=p_iter->get_entry();request+="=\'";
        do_escape(p_iter->get_pattern(),clean);
        request+=clean;
        request+="\'";
    }
    request+=MessageArray[2];
    //    request+= " WHERE ";
    for (p_iter = identifiers->begin();p_iter != identifiers->end();++p_iter) {
        if (p_iter != identifiers->begin())
            request += " AND ";
        request+=p_iter->get_entry();request+=" = \'";
        do_escape(p_iter->get_pattern(),clean);
        request+=clean;
        request+="\'";
    }
    do_message(request.c_str());
    i = mysql_query(Sql_Connection,request.c_str());
    if (i == -1) {
        //if some goes wrong 
        do_message(get_last_error());
        return i;
    }
    //should be >= 0 
    i = mysql_affected_rows(Sql_Connection);
    //log it
    do_message(FormatSTLString("%i %s",i,MessageArray[0]));
    return i;
}

int CMySql::delete_row(const char*db,const char*table,
                       PatternList*identifiers)
{
    std::string query;
    std::string clean;
    int i;
    unsigned long zaehler;

    if (!Sql_Connection){
        do_message(MessageArray[4]);
        return -1;
    }
    if (!table || strlen(table) == 0) {
        do_message(MessageArray[3]);
        return -1;
    }
    if (!db) {
	do_message(MessageArray[6]);
	return -1;
    }
    
    i = connect_db(db);
    if (i < 0)
        return i;

    query = "DELETE FROM ";
    //query+=db;query+='.';
    query+= table;
    PatternIterator p_iter;
    if (identifiers && identifiers->size() > 0) {
        query+=" ";
        query+=MessageArray[2];
        for (p_iter=identifiers->begin();p_iter != identifiers->end();++p_iter) {
            if (p_iter != identifiers->begin())
                query+=" AND ";
            query+=p_iter->get_entry();query+=" = \'";
            do_escape(p_iter->get_pattern(),clean);
            if (!clean.size()) {
                do_message("error escaping strings");
                return -1;
            }
            query+=clean;
            query+="\'";
        }    
    }
    do_message(query);
    i = mysql_query(Sql_Connection,query.c_str());
    if (i == -1) {
        //if some goes wrong 
        do_message(get_last_error());
        return i;
    }
    //should be > 0 
    zaehler = mysql_affected_rows(Sql_Connection);
    //log it
    do_message(FormatSTLString("%lu %s",zaehler,MessageArray[0]));
    return zaehler;
}


int CMySql::connect_db(const char*db)const
{
    int i = 0;
    if (!Sql_Connection) {
        do_message(MessageArray[4]);
        return -1;
    }
    if (!db)
        return i;
    i = mysql_select_db(Sql_Connection,db);
    if (i < 0) {
        do_message(get_last_error());
    }
    return i;
}

void CMySql::scramble_password(std::string&pass)const
{
    char Buffer[256];
    memset(Buffer,'\0',256);
    make_scrambled_password(Buffer,pass.c_str());
    pass = Buffer;
}

int CMySql::shutdown()
{
    int i;
    if (!Sql_Connection)
        return 0;
    i = mysql_shutdown(Sql_Connection);
    if (i < 0) {
        do_message(get_last_error());
    }
    return i;
}

bool CMySql::delete_table(const char*db, const char*table)
{
    std::string query;

    if (!db || !table) {
        do_message("delete_table: To few parameters");
        return false;
    }
    if (!Sql_Connection) {
        do_message(MessageArray[4]);
        return false;
    }
    if (!db) {
	do_message(MessageArray[6]);
	return false;
    }
    query = "";query = "DROP TABLE ";query+= db;query+='.';query+= table;
    do_message(query.c_str());
    int i = mysql_query(Sql_Connection,query.c_str());
    if (i < 0) {
        do_message(get_last_error());
        return false;
    }
    unsigned long zaehler = mysql_affected_rows(Sql_Connection);
    do_message(FormatSTLString("%lu %s",zaehler,MessageArray[0]));
    return true;
}

bool CMySql::get_table_keys(const char*db, const char*table, KeyList&target)const
{
    std::string command = "SHOW INDEX FROM ";
    std::string name="";
    MYSQL_RES*Res=0;
    MYSQL_ROW row;
	
    target.resize(0);
    if (!Sql_Connection) {
	do_message(MessageArray[4]);
	return false;
    }
    if (!table || strlen(table) == 0) {
	return false;
    }
    if (connect_db(db) < 0) {
	return false;
    }
    if (!db) {
	do_message(MessageArray[6]);
	return false;
    }
    //command+=db;command+='.';
    command+=table;
    int i = mysql_query(Sql_Connection,command.c_str());
    if (i < 0) {
	do_message(get_last_error());
	return false;
    }
    Res = mysql_store_result(Sql_Connection);
    
    if (!Res) {
        //autch - this should never be...
        command = mysql_error(Sql_Connection);
        if (!command.empty())
            do_message(command.c_str());
        else
            do_message("Error fetch the result");
        return false;
    }
    bool found = false;
    unsigned int count;
    while ((row=mysql_fetch_row(Res))) {
	for (count = 0; count < target.size(); ++count) {
	    if (target[count] == row[2]) {
		found = true;
		break;
	    }
	}
	if (!found) {
	    CKeyDescriptor key;
	    key.set_name((char*)row[2]);
	    if (strcmp((char*)row[1],"1") == 0) {
		key.setType(CKeyDescriptor::INDEX);
	    } else {
		if (strcasecmp((char*)row[2],"PRIMARY")==0) {
		    key.setType(CKeyDescriptor::PRIMARY);
		} else {
		    key.setType(CKeyDescriptor::UNIQUE);
		}
	    }
	    target.push_back(key);
	} else {
	    found = false;
	}
	name = (const char*)row[4];
	target[count].add_column_name(name,row[7]?atoi((const char*)row[7]):0);
	target[count].set_compound(target[count].column_count()>1);
    }
    mysql_free_result(Res);
    return true;
}

bool CMySql::rename_table(const char*db,const char*table,const char*new_name)const
{
    std::string command = "";
    if (!Sql_Connection) {
	do_message(MessageArray[4]);
	return false;
    }
    if (!table || strlen(table) == 0) {
	return false;
    }
    if (!db) {
	do_message(MessageArray[6]);
	return false;
    }
    if (connect_db(db) < 0) {
	return false;
    }
    if (is_system_db(db)) {
	do_message(MessageArray[5]);
	return false;
    }
    command = "ALTER TABLE ";
    command+= table;command+= " RENAME AS ";command+= new_name;
    do_message(command);
    int i = mysql_query(Sql_Connection,command.c_str());
    if (i < 0) {
	do_message(get_last_error());
	return false;
    }    
    return true;
}

bool CMySql::alter_table(const char*db, const char*table,ColumnList&n_columns,const KeyList&n_keys,
			 change_list&columns_change,KeyList&key_orig)const
{
    std::string command = "";
    std::string temp_command = "";
    unsigned int idx;
    change_list::iterator change_iterator;
    change_type temp_change;
	
    if (!Sql_Connection) {
	do_message(MessageArray[4]);
	return false;	
    }
    if (!table || strlen(table) == 0) {
	return false;
    }
    if (!db) {
	do_message(MessageArray[6]);
	return false;
    }
    if (is_system_db(db)) {
	do_message(MessageArray[5]);
	return false;
    }
    if (connect_db(db)!=0) {
        do_message("Could not connect to database");
		return false;
    }
    command = "ALTER TABLE ";
    //command+=db;command+='.';
    command+= table;

    int first_statement = 1;

    //looking for the primary key
    for (idx = 0; idx < key_orig.size();++idx) {
	if (key_orig[idx].getType() == CKeyDescriptor::PRIMARY) {
	    command+=" DROP PRIMARY KEY ";
	    first_statement = 0;
	}
    }

    for (idx = 0;idx < key_orig.size();++idx) {
	if (key_orig[idx].getType() != CKeyDescriptor::PRIMARY) {
	    if (!first_statement) {
		command+=", ";
	    } else {
		command+=" ";
		first_statement = 0;
	    }
	    command+="DROP KEY ";
	    command+=key_orig[idx].get_name();
	    command+=" ";
	}
    }
    for (idx = 0;idx < columns_change.size();++idx) {
	if (columns_change[idx].typ == change_type::del && 
	    columns_change[idx].columns.first.size()>0) {
	    if (!first_statement) {
		command+=", ";
	    } else {
		command+=" ";
		first_statement = 0;
	    }
	    command+="DROP ";
	    command+=columns_change[idx].columns.first;
	    command+=" ";
	}
    }
    for (idx = 0; idx < n_columns.size();++idx) {
	change_iterator = columns_change.find_second(n_columns[idx].get_name());
	if (change_iterator) {
	    if (!first_statement) {
		command+=", ";
	    } else {
		command+=" ";
		first_statement = 0;
	    }
	    switch (change_iterator->typ) {
	    case change_type::add:
		command+="ADD ";
		break;
	    case change_type::mod:
		command+="CHANGE ";
		command+=change_iterator->columns.first;
		command+=" ";
		break;
	    default:
		continue;
		break;
	    }
	    if (!build_command(temp_command,n_columns[idx])) {
		return false;
	    }
	    command+=temp_command;
	}
    }
    build_key_command(temp_command,&n_keys,true);
    if (temp_command.size()) 
	command+=temp_command;
    do_message(command);
    first_statement = mysql_query(Sql_Connection,command.c_str());
    if (first_statement < 0) {
	do_message(get_last_error());
	return false;
    }
    do_message(mysql_info(Sql_Connection));
    return true;
}

int CMySql::build_command(std::string&target,CColumnDescriptor&desc)const
{
    std::string Type = "";
    target.resize(0);
    target=desc.get_name();
    target+=" ";
    Type = TypeDescript(desc.get_FieldType());
    if (Type.length() ==  0) {
	do_message("Unknown fieldtype");
	return 0;
    }
    //special cases for enums and sets
	
    if (desc.get_FieldType() == 21 || desc.get_FieldType() == 22) {
	stringIlist current_enums;
	current_enums = desc.enum_list();
	if (current_enums.size() == 0) {
	    do_message("Empty SETs/ENUMs are not allowed!");
	    return 0;
	}
	unsigned int max_enum = 64;
	if (desc.get_FieldType()==17) {
	    max_enum = 65535;
	}
	if (current_enums.size() > max_enum) {
	    do_message(FormatSTLString("More than %u %ss are not allowed",max_enum,max_enum==64?"SET":"ENUM"));
	    return 0;
	}
	Type+="(";
	for (max_enum=0;max_enum<current_enums.size();++max_enum) {
	    if (max_enum)
		Type+=",";
	    Type+="'";
	    Type+=current_enums[max_enum];
	    Type+="'";
	}
	Type+=")";
    }
    target+=Type;
    target+=" ";
    if (desc.get_MaxLength()>0 && desc.get_FieldType() != 21 && desc.get_FieldType() != 22) {
	target+="(";
	target+=FormatSTLString("%i",desc.get_MaxLength());
	if (desc.get_MaxDecimals() > 0) {
	    target+=", ";
	    target+=FormatSTLString("%i",desc.get_MaxDecimals());
	}
	target+=") ";
    }
    if (desc.is_Unsigned())
	target+="UNSIGNED ";
    if (desc.is_ZeroFill())
	target+="ZEROFILL ";
    if (desc.is_Binary()) 
	target+="BINARY ";
    if (desc.is_NotNull()) {
	target+="NOT NULL ";
	if (desc.get_default() && strlen(desc.get_default())>0) {
	    target+="DEFAULT '";
	    target+=desc.get_default();
	    target+="' ";
	}
    }
    if (desc.is_AutoIncrement())
	target+="AUTO_INCREMENT ";
    return 1;
}

int CMySql::build_key_command(std::string&target,const KeyList*keys,bool add)
{
    target.resize(0);
    KeyList::const_iterator key;
    if (keys) {
        int already_prim = 0;
        pairIlist col_list;
        
        int f_count = 0;
        for (key = keys->begin(); key != keys->end(); ++key) {
            f_count = 0;
            if (add) 
		target+=", ADD ";
	    else
		target+=", ";
            switch(key->getType()) {
            case CKeyDescriptor::PRIMARY:
                if (!already_prim) {
                    target+="PRIMARY KEY ";
                    already_prim = 1;
                } else {
                    target+="UNIQUE ";
                    //key->setType(CKeyDescriptor::UNIQUE);
                }
                break;
            case CKeyDescriptor::UNIQUE:
                target+="UNIQUE ";
                break;
            case CKeyDescriptor::INDEX:
                target+="INDEX ";
                break;
            default:
                continue;
            }
            if (key->getType() != CKeyDescriptor::PRIMARY) {
                target+=key->get_name();
            }
            col_list = key->get_ColList();
            if (col_list.size() == 0)
                continue;
            target+="(";
            pairIlist::iterator iter;
            iter = col_list.begin();
            for (;iter!=col_list.end();++iter) {
                if (f_count> 0)
                    target+=",";
                ++f_count;
                target+=iter->first.c_str();
                if (iter->second > 0) {
                    target+="(";
                    target+=FormatSTLString("%u",iter->second);
                    target+=")";
                }
            }
            target+=") ";
        }
    }
    return 1;
}

bool CMySql::do_statement(const char*statement,stringlist&target,stringlist&sizes,stringlist&headers,
			  bool log_errors_only)
{
    sizes.resize(0);
    target.resize(0);

    if (!Sql_Connection) {
	do_message(MessageArray[4]);
	return false;	
    }
    MYSQL_ROW row;
    MYSQL_FIELD*Field;

    std::string query;

    MYSQL_RES*Res=0;
    if (!log_errors_only)
	do_message(statement);
    int result = mysql_query(Sql_Connection,statement);
    if (result < 0) {
	if (log_errors_only)
	    do_message(statement);
	do_message(get_last_error());
	return false;
    }
    Res = mysql_store_result(Sql_Connection);
    if (!Res) {
        query = mysql_error(Sql_Connection);
        if (!query.empty()) {
	    if (log_errors_only)
		do_message(statement);
            do_message(query.c_str());
	    return false;
	}
        return true;
    }
    unsigned long zaehler = Res->row_count;
    if (!log_errors_only)
	do_message(FormatSTLString("%lu %s",zaehler,MessageArray[0]));
    std::string helper;
    std::string entry;
    unsigned int j;
    while ((row=mysql_fetch_row(Res))) {
        for (j = 0;j < mysql_num_fields(Res);j++) {     
	    query.resize(0);
            query = row[j]?(char*)row[j]:"";
            //how long is the longest?
            Res->current_field = j;
            Field = mysql_fetch_field(Res);
            // we retrieve the first line splitted, for request, which will only need this line
            // for instance a query to the user rights
	    unsigned int max_size = Field->max_length;
	    std::string f = Field->name;
	    if (f.size()< max_size) {
		f.resize(max_size,' ');
	    } else {
		max_size = f.size();
	    }
            if (target.size() == 0) {
		headers.push_back(std::string(f));
            }
            //ok - justify
	    int groesse = max_size-query.size();
            helper.resize(0);
	    if (IS_NUM(Field->type)) {
		helper.resize(groesse>=0?groesse:0,' ');
	    }
	    helper+=query;
	    if (!IS_NUM(Field->type)) {
		helper.resize(max_size,' ');
	    }
            if (helper.size() == 0)
                helper = " ";
            if (j<mysql_num_fields(Res)-1)
                // its the simplest way to select datas - hope there is no tab in a field
                helper+="\t";
            entry+=helper;
        }
        //append to the list
        target.push_back(entry);
        entry = "";
    }
    mysql_free_result(Res);
    return true;
}
